"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const error_1 = require("@pnpm/error");
const lockfile_file_1 = require("@pnpm/lockfile-file");
const outdated_1 = require("@pnpm/outdated");
const semver_diff_1 = require("@pnpm/semver-diff");
const store_path_1 = require("@pnpm/store-path");
const chalk_1 = require("chalk");
const R = require("ramda");
const stripColor = require("strip-color");
const table = require("text-table");
const createLatestVersionGetter_1 = require("../createLatestVersionGetter");
const readImporterManifest_1 = require("../readImporterManifest");
async function default_1(args, opts, command) {
    const packages = [
        {
            manifest: await readImporterManifest_1.readImporterManifestOnly(opts.prefix, opts),
            path: opts.prefix,
        },
    ];
    const { outdatedPackages } = (await outdatedDependenciesOfWorkspacePackages(packages, args, opts))[0];
    if (!outdatedPackages.length)
        return;
    const columnNames = [
        'Package',
        'Current',
        'Latest',
    ].map((txt) => chalk_1.default.underline(txt));
    let columnFns = [
        renderPackageName,
        renderCurrent,
        renderLatest,
    ];
    return table([
        columnNames,
        ...R.sortWith([
            sortBySemverChange,
            (o1, o2) => o1.packageName.localeCompare(o2.packageName),
        ], outdatedPackages.map(toOutdatedWithVersionDiff))
            .map((outdatedPkg) => columnFns.map((fn) => fn(outdatedPkg))),
    ], {
        stringLength: (s) => stripColor(s).length,
    });
}
exports.default = default_1;
function toOutdatedWithVersionDiff(outdated) {
    if (outdated.latest) {
        return Object.assign({}, outdated, semver_diff_1.default(outdated.wanted, outdated.latest));
    }
    return Object.assign({}, outdated, { change: 'unknown' });
}
exports.toOutdatedWithVersionDiff = toOutdatedWithVersionDiff;
function renderPackageName({ belongsTo, packageName }) {
    switch (belongsTo) {
        case 'devDependencies': return `${packageName} ${chalk_1.default.dim('(dev)')}`;
        case 'optionalDependencies': return `${packageName} ${chalk_1.default.dim('(optional)')}`;
        default: return packageName;
    }
}
exports.renderPackageName = renderPackageName;
function renderCurrent({ current, wanted }) {
    let output = current || 'missing';
    if (current === wanted)
        return output;
    return `${output} (wanted ${wanted})`;
}
exports.renderCurrent = renderCurrent;
const DIFF_COLORS = {
    feature: chalk_1.default.yellowBright.bold,
    fix: chalk_1.default.greenBright.bold,
};
function renderLatest({ latest, change, diff }) {
    if (!latest)
        return '';
    if (change === null || !diff)
        return latest;
    const highlight = DIFF_COLORS[change] || chalk_1.default.redBright.bold;
    const same = joinVersionTuples(diff[0], 0);
    const other = highlight(joinVersionTuples(diff[1], diff[0].length));
    if (!same)
        return other;
    return diff[0].length === 3 ? `${same}-${other}` : `${same}.${other}`;
}
exports.renderLatest = renderLatest;
function joinVersionTuples(versionTuples, startIndex) {
    const neededForSemver = 3 - startIndex;
    if (versionTuples.length <= neededForSemver)
        return versionTuples.join('.');
    return `${versionTuples.slice(0, neededForSemver).join('.')}-${versionTuples.slice(neededForSemver).join('.')}`;
}
function sortBySemverChange(outdated1, outdated2) {
    return pkgPriority(outdated1) - pkgPriority(outdated2);
}
exports.sortBySemverChange = sortBySemverChange;
function pkgPriority(pkg) {
    switch (pkg.change) {
        case null: return 0;
        case 'fix': return 1;
        case 'feature': return 2;
        case 'breaking': return 3;
        default: return 4;
    }
}
async function outdatedDependenciesOfWorkspacePackages(pkgs, args, opts) {
    const lockfileDirectory = opts.lockfileDirectory || opts.prefix;
    const currentLockfile = await lockfile_file_1.readCurrentLockfile(lockfileDirectory, { ignoreIncompatible: false });
    const wantedLockfile = await lockfile_file_1.readWantedLockfile(lockfileDirectory, { ignoreIncompatible: false }) || currentLockfile;
    if (!wantedLockfile) {
        throw new error_1.default('OUTDATED_NO_LOCKFILE', 'No lockfile in this directory. Run `pnpm install` to generate one.');
    }
    const store = await store_path_1.default(opts.prefix, opts.store);
    const getLatestVersion = createLatestVersionGetter_1.default(Object.assign({}, opts, { lockfileDirectory,
        store }));
    return Promise.all(pkgs.map(async ({ manifest, path }) => {
        const optsForOutdated = {
            currentLockfile,
            getLatestVersion,
            lockfileDirectory,
            manifest,
            prefix: path,
            wantedLockfile,
        };
        return {
            manifest,
            outdatedPackages: args.length
                ? await outdated_1.forPackages(args, optsForOutdated)
                : await outdated_1.default(optsForOutdated),
            prefix: lockfile_file_1.getLockfileImporterId(lockfileDirectory, path),
        };
    }));
}
exports.outdatedDependenciesOfWorkspacePackages = outdatedDependenciesOfWorkspacePackages;
